from .Broker import Broker
from .Strategy import Strategy
from .Bot import Bot
from .Notifier import Notifier, addMessage
from .Data import Data
from .CSVData import CSVData
from .BacktestBroker import BacktestBroker
from .LNMBroker import LNMBroker



from abc import ABCMeta, abstractmethod


class Strategy(metaclass=ABCMeta):

    def __init__(self, broker=None, **params):
        self.params = params
        self.broker = broker
        self.datas = []
        self.init()

    @abstractmethod
    def init(self):
        """
        This method is called only once, before any price bar is recieved.
        """
        pass

    @abstractmethod
    def execute(self):
        """
        This method is called for every new bar recieved.
        """
        pass

    @abstractmethod
    def stop(self):
        """
        This method is called only once, when there is no new price bars left.
        """
        pass

    def addBroker(self, broker):
        self.broker = broker

    def addData(self, data):
        self.datas.append(data)



import sys
import asyncio


def background(f):
    def wrapped(*args, **kwargs):
        return asyncio.get_event_loop().run_in_executor(None, f,
                                                        *args, **kwargs)

    return wrapped


class Bot:

    def __init__(self):
        self.strategy = []

    def run(self):
        for strat in self.strategy:
            self.runStrategy(strat)

    # @background
    def runStrategy(self, strategy):
        assert(strategy.datas)
        priceGens = []
        for data in strategy.datas:
            priceGens.append(data.dataGenerator())
        while True:
            prices = []
            try:
                for gen in priceGens:
                    prices.append(next(gen))
            except StopIteration:
                break
            try:
                strategy.broker.processData(prices)
                strategy.execute(prices)
            except ValueError as VErr:
                strategy.broker.notifier.notify(str(VErr))
                break
            except Exception as exception:
                strategy.broker.notifier.notify(str(exception))
                raise exception
            sys.stdout.flush()
        strategy.stop()

    def addStrategy(self, strategy):
        if strategy.broker is None:
            raise ValueError("Add Broker to Strategy first")
        self.strategy.append(strategy)



from .Broker import Broker
from .Notifier import addMessage
from LNMarkets import Positions


class LNMBroker(Broker):

    def __init__(self, token, initialBalance, silent=False):
        self.token = token
        self.initialBalance = initialBalance
        super().__init__()

    @staticmethod
    def calculateProfit(positions):
        totalProfit = 0.0
        for position in positions:
            totalProfit += float(position['pl'])

        return totalProfit

    @staticmethod
    def calculateMargin(positions):
        totalMargin = 0.0
        for position in positions:
            totalMargin += float(position['margin'])
        return totalMargin

    @property
    def position(self):
        pass

    @property
    def openPositions(self):
        return Positions.getPositions(self.token, "open")['open']

    @property
    def closedPositions(self):
        return Positions.getPositions(self.token, "closed")['closed']

    @property
    def unrealizedProfit(self):
        return self.calculateProfit(self.openPositions)

    @property
    def realizedProfit(self):
        return self.calculateProfit(self.closedPositions)

    @property
    def balance(self):
        return self.cashBalance + self.unrealizedProfit

    @property
    def cashBalance(self):
        return self.realizedProfit + self.initialBalance

    @property
    def marginWithheld(self):
        return self.calculateMargin(self.openPositions)

    @addMessage
    def buy(self, leverage, quantity, stoploss=None, takeprofit=None,
            limit=None):
        if limit is None:
            return Positions.buy(
                token=self.token,
                leverage=leverage,
                quantity=quantity,
                stoploss=stoploss,
                takeprofit=takeprofit,
            )
        else:
            return Positions.limitBuy(
                token=self.token,
                leverage=leverage,
                price=limit,
                quantity=quantity,
                stoploss=stoploss,
                takeprofit=takeprofit,
            )

    @addMessage
    def sell(self, leverage, quantity, stoploss=None, takeprofit=None,
             limit=None):
        if limit is None:
            return Positions.sell(
                token=self.token,
                leverage=leverage,
                quantity=quantity,
                stoploss=stoploss,
                takeprofit=takeprofit,
            )
        else:
            return Positions.limitSell(
                token=self.token,
                leverage=leverage,
                price=limit,
                quantity=quantity,
                stoploss=stoploss,
                takeprofit=takeprofit,
            )

    @addMessage
    def updateProfit(self, pid, price):
        return Positions.updatePosition(
            token=self.token,
            pid=pid,
            type_='takeprofit',
            value=price,
            )

    @addMessage
    def updateStoploss(self, pid, price):
        return Positions.updatePosition(
            token=self.token,
            pid=pid,
            type_='stoploss',
            value=price,
            )

    @addMessage
    def closePosition(self, pid):
        return Positions.closePosition(self.token, pid)

    @addMessage
    def closeAllLongs(self):
        return Positions.closeAllLongs(self.token)

    @addMessage
    def closeAllShorts(self):
        return Positions.closeAllShorts(self.token)

    @addMessage
    def cancelPosition(self, pid):
        return Positions.cancelPosition(self.token, pid)

    @addMessage
    def addMargin(self, pid, amount):
        return Positions.addMargin(self.token, pid, amount)

    @addMessage
    def cashin(self, pid, amount):
        return Positions.cashin(self.token, pid, amount)

    def processData(self, priceData):
        pass



import asyncio
import time
from .Data import Data
import krakenex
from pykrakenapi import KrakenAPI


class KrakenData(Data):

    def __init__(self, interval=1, **params):
        self.KAPI = KrakenAPI(krakenex.API())
        self.ohlc, self.since = self.KAPI.get_ohlc_data(
            "BTCUSD",
            interval=interval,
            since=None,
            ascending=True,
        )
        self.firstCall = True
        super().__init__(interval=interval, **params)

        def dataGenerator(self):
            assert(self.liveTrading)
            wait = 60
            maxRows = 2880
            blockSize = 1440
            if not self.firstCall:
                # await asyncio.sleep(interval*wait)
                time.sleep(interval*wait)
            else:
                self.firstCall = False

            for retry in range(10):
                try:
                    ohlcnew, self.since = self.KAPI.get_ohlc_data(
                        "BTCUSD",
                        interval=interval,
                        since=self.since,
                        ascending=True,
                    )
                except (TimeoutError, ConnectionError):
                    # await asyncio.sleep(wait)
                    time.sleep(wait)
                    retry += 1
                    continue
                # Remove the last enty as it is unconfirmed
                self.ohlc.drop(self.ohlc.tail(1).index, inplace=True)
                self.ohlc = self.ohlc.append(ohlcnew)
                if len(self.ohlc) >= maxRows:
                    self.ohlc.drop(self.ohlc.head(blockSize).index,
                                   inplace=True)
                    yield self.ohlc



from abc import ABCMeta, abstractmethod


class Data(metaclass=ABCMeta):

    def __init__(self):
        pass

    @abstractmethod
    def dataGenerator(self):
        """
        Genrtator which yields datetime index dataframe containing
        open, high, low, and close prices with volume
        for each new price bar recieved.
        """
        pass
